c -*- mode: fundamental -*-
c =================================================================
c OMP : pgf90 -V -Mmp -o pi_omp.exe pi_omp.f
c
c compute pi by integrating f (x) = 4/(1 + x**2) from 0 to 1
c
c Ref.: Koelbel C. H., Loveman D.B., Schreiber R.S., 
c       Steele G.L., Zosel M.E.,
c       The High Perfomance Fortran Handbook,
c       fig. 3.1, pag. 86,
c       The Mit Press, 2nd edition, 1997.
c -----------------------------------------------------------------
        program  pi_omp
        implicit none
        integer, parameter :: idp = kind (1.0 d0)
        integer, parameter :: iin = kind (1)
c .................................................................
        real    (idp), parameter ::   cero = 0.0 d0
        real    (idp), parameter ::   half = 0.5 d0
        real    (idp), parameter ::    uno = 1.0 d0
        real    (idp), parameter :: cuatro = 4.0 d0
c .................................................................
        integer (iin), parameter :: n = 11 000   ! recta
        real    (idp), parameter :: h = uno / n  ! ancho 
c .................................................................
        integer       :: omp_get_thread_num
        integer       :: np
        real    (idp) :: pi
        real    (idp) :: x, sum
        integer (iin) :: i
c -----------------------------------------------------------------
        np = omp_get_thread_num ()
        write (*,*)
        write (*,*)' nro_de_procesos ; np = ', np
        sum = cero
c$omp   parallel
c$omp&  default (shared), private (x)
c$omp&  reduction (+: sum)
c$omp&  do
        do i = 1, n
          x = h * (i - half)
          sum = sum + cuatro / (uno + x * x)
        end do
c$omp   end parallel
        pi = h * sum
        write (*,*)
        write (*,*)' pi = ', pi
        write (*,*)
        end program
c =================================================================
