c -*- mode: fundamental -*-
c ==================================================================
c  pgf90 -lfmpich -lmpich -c mpi.hpf  
c  pgf90 -lfmpich -lmpich -c vida1.hpf
c  pgf90 -lfmpich -lmpich -o vida1.exe mpi.o vida1.o
c
c  mpirun -machinefile machi.dat -nolocal -np 1 vida1.exe
c ------------------------------------------------------------------
c This program simulates the game of life cellular automaton
c using mpi to distribute the calculations to parallel tasks.
c
c MPI version
c
c Ref. OpenMP course, v. 1.1, Graham P.
c      Edinburgh Parallel Computing Centre, 1999
c ------------------------------------------------------------------
        program vida1
        use  mpi
        implicit none
c ------------------------------------------------------------------
c   precision "q" reducida: -10**q .le. n .le. +10**q
c ------------------------------------------------------------------
        integer, parameter :: vcor0 = selected_int_kind (0)
        integer, parameter :: vcor1 = selected_int_kind (1)
        integer, parameter :: vcor2 = selected_int_kind (2)
        integer, parameter :: vcor3 = selected_int_kind (3)
        integer, parameter :: vcor4 = selected_int_kind (4)
c ------------------------------------------------------------------
c   parametros de la corrida
c ------------------------------------------------------------------
        integer, parameter :: n_generaciones =  100
        integer, parameter ::          xsiza =   30 ! tablero
        integer, parameter ::          ysiza =   30 ! tablero
        integer, parameter ::        maxgrey =    1 ! celula viva
        integer, parameter ::              h =   10 ! patron inicial
        integer, parameter ::           cmax =   70 ! max col.impr.
        integer, parameter ::       print_id =    0 ! 1 proc. impr.
        logical, parameter ::        imprime = .true.
c ------------------------------------------------------------------
c   variables de informacion basica
c ------------------------------------------------------------------
        integer  :: err, rank
        integer  :: siza (2)
c ------------------------------------------------------------------
c   variables para crear dimensiones y topologia cartesiana
c ------------------------------------------------------------------
        integer , parameter :: ndims = 2
        integer  :: dims (2)
        integer  :: twodim_comm
        logical  :: periods (2), reorder
        integer  :: nx, ny, xglob, yglob
c ------------------------------------------------------------------
c   variables para tipo de datos derivados
c ------------------------------------------------------------------
        integer  :: nblocks, lblocks, memtype
        integer  :: stride, subdom, row, column
        integer  :: subsiza (2), starts (2)
c ------------------------------------------------------------------
c   matrices de trabajo
c ------------------------------------------------------------------
        integer  :: tablero (0:xsiza+1,0:ysiza+1)
        integer  :: vecinos (xsiza, ysiza)
c ------------------------------------------------------------------
c   variable para la posicion de los procesadores
c ------------------------------------------------------------------
        integer  :: coords (ndims)
c ------------------------------------------------------------------
c   variable para los desplazamientos
c ------------------------------------------------------------------
        integer  :: direction, disp 
        integer  :: vecino_izquie, vecino_derech
        integer  :: vecino_arriba, vecino_debajo
c ------------------------------------------------------------------
c   variables para enviar y recibir
c ------------------------------------------------------------------
        integer  :: count, tag, dx, dy
        integer  :: request (4)
        integer  :: status (mpi_status_size,4)
        integer  :: n_vivas
        integer  :: i_vivas
c ------------------------------------------------------------------
c   variables extras
c ------------------------------------------------------------------
        integer  :: i, j, m, n, p
        integer  :: my_id
        integer  :: comm
        integer  :: ntasks
c ------------------------------------------------------------------
c   variables para cronometro
c ------------------------------------------------------------------
        real   , dimension (3) :: time
        integer                :: h1, h2, h3
        real   , parameter     :: cero = 0.0
c -----------------------------------------------------------------
c   inicia cronometro;  t1 = second ()
c -----------------------------------------------------------------
        time = cero
        call system_clock (count = h1, count_rate = h3)
c ------------------------------------------------------------------
c   inicializa mpi
c ------------------------------------------------------------------
        call mpi_init      (err)
        call mpi_comm_rank (mpi_comm_world, rank, err)
        call mpi_comm_size (mpi_comm_world, siza, err)
        ntasks = siza (1)
        my_id  = rank
c ------------------------------------------------------------------
c   crea dimensiones
c ------------------------------------------------------------------
        dims (1) = 0
        dims (2) = 0
        call  mpi_dims_create (siza, ndims, dims, err)
        nx = dims (1)
        ny = dims (2)
        dx = xsiza / nx
        dy = ysiza / ny
c ------------------------------------------------------------------
c   crea una grilla cartesiana 2D periodica
c ------------------------------------------------------------------
        do  i = 1, ndims
        periods (i) = .true.
        end do ! i
        reorder = .true.
        call  mpi_cart_create (mpi_comm_world, 
     #                         ndims, 
     #                         dims, 
     #                         periods, 
     #                         reorder, 
     #                         twodim_comm, 
     #                         err)
c ------------------------------------------------------------------
c   determina las coordenadas cartesianas del procesador (x,y)
c ------------------------------------------------------------------
        call  mpi_cart_coords (twodim_comm, 
     #                         rank, 
     #                         ndims, 
     #                         coords, 
     #                         err)
c ------------------------------------------------------------------
c   inicializa tablero
c ------------------------------------------------------------------
        xglob = coords (1) * dx
        yglob = coords (2) * dy
        tablero = 0
        m = xsiza / 2
        do  i = m - h/2 + 1, m + h/2
        tablero (i,m) = maxgrey 
        end do ! i
        n = maxval ( (/ xsiza, ysiza /) )
c ------------------------------------------------------------------
c   crea datos derivados
c ------------------------------------------------------------------
        siza (1) = xsiza
        siza (2) = ysiza
        subsiza (1) = dx
        subsiza (2) = dy
        starts  (1) = coords (1) * dx
        starts  (2) = coords (2) * dy
        call  mpi_type_create_subarray (2, 
     #                                  siza, 
     #                                  subsiza, 
     #                                  starts,
     #                                  mpi_order_fortran,
     #                                  mpi_integer,
     #                                  subdom, 
     #                                  err)
c ------------------------------------------------------------------
c   esquema para la memoria
c ------------------------------------------------------------------
        siza (1) = dx + 2
        siza (2) = dy + 2
        starts (1) = 1
        starts (2) = 1
        call  mpi_type_create_subarray (2, 
     #                                  siza, 
     #                                  subsiza, 
     #                                  starts,
     #                                  mpi_order_fortran,
     #                                  mpi_integer,
     #                                  memtype,
     #                                  err)
        call mpi_type_commit (memtype, err)
c ------------------------------------------------------------------
c   fila
c ------------------------------------------------------------------
        nblocks = dy + 2
        lblocks = 1
        stride  = xsiza + 2
        call  mpi_type_vector (nblocks, 
     #                         lblocks, 
     #                         stride,
     #                         mpi_integer,
     #                         row,
     #                         err)
        call mpi_type_commit  (row, err)
c ------------------------------------------------------------------
c   columna
c ------------------------------------------------------------------
        nblocks = 1
        lblocks = dx + 2
        stride  = xsiza + 2
        call  mpi_type_vector (nblocks, 
     #                         lblocks, 
     #                         stride,
     #                         mpi_integer,
     #                         column,
     #                         err)
        call mpi_type_commit  (column, err)
c ------------------------------------------------------------------
c   busca el vecino mas cercano en la direccion x
c ------------------------------------------------------------------
        direction = 1
        disp      = 1
        call  mpi_cart_shift  (twodim_comm, 
     #                         direction, 
     #                         disp,
     #                         vecino_izquie, 
     #                         vecino_derech, 
     #                         err)
c ------------------------------------------------------------------
c   busca el vecino mas cercano en la direccion y
c ------------------------------------------------------------------
        direction = 0
        disp      = 1
        call  mpi_cart_shift (twodim_comm, 
     #                        direction, 
     #                        disp,
     #                        vecino_arriba, 
     #                        vecino_debajo, 
     #                        err)
        count = 1
        tag   = 1
c ------------------------------------------------------------------
c   tablero: impresion eventual
c ------------------------------------------------------------------
        if (my_id == print_id .and. imprime .and. n .le. cmax) then
          open  (1, file = 'vida_mpi.map', status = 'unknown')
          close (1, status = 'delete') 
        end if
c ------------------------------------------------------------------
c   ajusta cronometro
c ------------------------------------------------------------------
        if  ( my_id == print_id ) then
          call system_clock (count = h2)     ! t2 = second ()
          time (1) = h2 - h1                 ! time (1) = t2 - t1
          call system_clock (count = h1)     ! t1 = second ()
        end if
c ------------------------------------------------------------------
c   titulos
c ------------------------------------------------------------------
        if  ( my_id == print_id ) then
          write (*,*)
          write (*,*)' juego de la vida con MPI '
          write (*,*)' number_of_processors = ', ntasks
        end if
c ------------------------------------------------------------------
c    lazo sobre las generaciones
c ------------------------------------------------------------------
        do   p = 1, n_generaciones
c ------------------------------------------------------------------
c   tablero: impresion eventual
c ------------------------------------------------------------------
        if (my_id == print_id .and. imprime .and. n .le. cmax) then
          open  (1, file = 'vida_mpi.map', status = 'unknown')
          do
          read  (1,*, end = 40)
          end do
   40     write (1,'(i5)') p
          do     i = 1, xsiza
          do     j = 1, ysiza
          if ( tablero (i,j) .eq. 1 ) then
            write (1,80,advance='no') '*'
          else
            write (1,80,advance='no') '.'
          end if
          end do ! j
          write (1,*,advance='yes')
          end do ! i
          close (1, status = 'keep') 
        end if
c ------------------------------------------------------------------
c    permuta bordes con el procesador, arriba y abajo
c ------------------------------------------------------------------
        call  mpi_issend  ( tablero (dx,0), 
     #                      count, 
     #                      row, 
     #                      vecino_debajo,
     #                      tag,
     #                      twodim_comm,
     #                      request (1),
     #                      err)
        call  mpi_issend  ( tablero (1,0), 
     #                      count, 
     #                      row, 
     #                      vecino_arriba,
     #                      tag,
     #                      twodim_comm,
     #                      request (2),
     #                      err)
        call  mpi_irecv   ( tablero (0,0), 
     #                      count, 
     #                      row, 
     #                      vecino_arriba,
     #                      tag,
     #                      twodim_comm,
     #                      request (3),
     #                      err)
        call  mpi_irecv   ( tablero (dx+1,0), 
     #                      count, 
     #                      row, 
     #                      vecino_debajo,
     #                      tag,
     #                      twodim_comm,
     #                      request (4),
     #                      err)
        call  mpi_waitall ( 4, request, status, err)
c ------------------------------------------------------------------
c    permuta bordes con el procesador, izquierda y derecha
c ------------------------------------------------------------------
        call  mpi_issend  ( tablero (0,1), 
     #                      count, 
     #                      column,
     #                      vecino_izquie,
     #                      tag,
     #                      twodim_comm,
     #                      request (1),
     #                      err)
        call  mpi_issend  ( tablero (0,dy), 
     #                      count, 
     #                      column, 
     #                      vecino_derech,
     #                      tag,
     #                      twodim_comm,
     #                      request (2),
     #                      err)
        call  mpi_irecv   ( tablero (0,dy+1), 
     #                      count, 
     #                      column, 
     #                      vecino_derech,
     #                      tag,
     #                      twodim_comm,
     #                      request (3),
     #                      err)
        call  mpi_irecv   ( tablero (0,0), 
     #                      count, 
     #                      column, 
     #                      vecino_izquie,
     #                      tag,
     #                      twodim_comm,
     #                      request (4),
     #                      err)
        call  mpi_waitall ( 4, request, status, err)
c ------------------------------------------------------------------
c    lazo sbre las celdas locales para contar los vecinos
c ------------------------------------------------------------------
         do  j = 1, dy
         do  i = 1, dx
         vecinos (i,j) = 
     #      (   tablero (i-1,j+1) + tablero (  i,j+1)
     #        + tablero (i+1,j+1) + tablero (i-1,  j)
     #        + tablero (i+1,  j) + tablero (i-1,j-1)
     #        + tablero (  i,j-1) + tablero (i+1,j-1) ) / maxgrey
         end do ! i
         end do ! j
c ------------------------------------------------------------------
c    lazo sbre las celdas locales, actualiza vivas en el tablero
c ------------------------------------------------------------------
         i_vivas = 0
         do  j = 1, dy
         do  i = 1, dx
         if     (vecinos (i,j) .lt. 2 .or. 
     #           vecinos (i,j) .gt. 3 ) then
                 tablero (i,j) = 0 
         elseif (vecinos (i,j) .eq. 3 ) then
                 tablero (i,j) = maxgrey
         end if
         if  ( tablero (i,j) .eq. 1 ) i_vivas = i_vivas + 1
         end do ! i
         end do ! j
c ------------------------------------------------------------------
c    operacion de reduccion: celulas vivas en todo el tablero
c ------------------------------------------------------------------
         call  mpi_reduce (i_vivas, 
     #                     n_vivas, 
     #                     1, 
     #                     mpi_integer,
     #                     mpi_sum,
     #                     0,
     #                     twodim_comm)
c ------------------------------------------------------------------
c   pasa a la siguiente generacion
c ------------------------------------------------------------------
         end do ! p
c ------------------------------------------------------------------
c   end the mpi run
c ------------------------------------------------------------------
         call  mpi_finalize
c ------------------------------------------------------------------
c   ajusta cronometro
c ------------------------------------------------------------------
        if  ( my_id == print_id ) then
          call system_clock (count = h2)     ! t2 = second ()
          time (2) = h2 - h1                 ! time (2) = t2 - t1
          time (3) = time (1) + time (2)     ! total_time
          time = time / h3                   ! pasa de ciclos a seg
        end if
c ------------------------------------------------------------------
c   nro de celulas vivas final
c ------------------------------------------------------------------
        if  ( my_id == print_id ) then
          n_vivas = 0
          n_vivas = sum ( tablero, mask = tablero .eq. 1)
          write (*,*)' vivas = ', n_vivas
        end if
c ------------------------------------------------------------------
c   imprime
c ------------------------------------------------------------------
        if  ( my_id == print_id ) then
          write (*,*)
          write (*,*)' tpo de inicial.  [s] ; t1 = ', real (time(1))
          write (*,*)' tpo de barrido   [s] ; t2 = ', real (time(2))
          write (*,*)' tiempo_total     [s] ; t3 = ', real (time(3))
          write (*,*)
        end if
c ------------------------------------------------------------------
c   imprime en archivo historico
c ------------------------------------------------------------------
        if  ( my_id == print_id ) then
          open (1, file = 'vida_mpi.tpo', status = 'unknown')
          do
          read (1,*, end = 50)  i
          end do
   50     write  (1,60)  ntasks, time (1:3)
          close  (1, status = 'keep')
         end if
c ------------------------------------------------------------------
c   codigos de formato
c ------------------------------------------------------------------
   60   format (1x, i4, 3 (1x, e19.9) )
   70   format (i1)
   80   format (a1)
c ..................................................................
         end program
c ==================================================================
