c -*- mode: fundamental -*-
c ==================================================================
c  pghpf -V -fast -Mautopar -Mmp -c vida3.hpf
c  pghpf -V -fast -Mautopar -Mmp -o vida3.exe vida3.o
c
c  pghpf -V -fast -Mautopar -Mmp -o vida3.exe vida3.hpf 
c
c  vida3.exe  -pghpf -machinefile machi.dat -np 8
c ------------------------------------------------------------------
c This program simulates the game of life cellular automaton
c using mpi to distribute the calculations to parallel tasks.
c
c OMP version
c
c Ref. OpenMP course, v. 1.1, Graham P.
c      Edinburgh Parallel Computing Centre, 1999
c ------------------------------------------------------------------
        program vida3
        implicit none
c ------------------------------------------------------------------
c   parametros de la corrida
c ------------------------------------------------------------------
        integer, parameter :: n_generaciones =  100
        integer, parameter ::              n =   30 ! tablero
        integer, parameter ::              h =   10 ! patron inicial
        integer, parameter ::           cmax =   70 ! max col.impr.
        logical, parameter ::        imprime = .true.
c ------------------------------------------------------------------
c   precision "q" reducida: -10**q .le. n .le. +10**q
c ------------------------------------------------------------------
        integer, parameter :: vcor0 = selected_int_kind (0)
        integer, parameter :: vcor1 = selected_int_kind (1)
c ------------------------------------------------------------------
c   declara matrices
c ------------------------------------------------------------------
        integer (vcor0), dimension (0:n+1,0:n+1) :: tablero
        integer (vcor1), dimension (0:n+1,0:n+1) :: vecinos
c ------------------------------------------------------------------
c   otras variables
c ------------------------------------------------------------------
        integer :: i, j, p, m 
        integer :: n_vivas, ntasks
        integer :: omp_get_num_threads
c ------------------------------------------------------------------
c   variables para cronometro
c ------------------------------------------------------------------
        real   , dimension (3) :: time
        integer                :: h1, h2, h3
        real   , parameter     :: cero = 0.0
c ------------------------------------------------------------------
c   titulos
c ------------------------------------------------------------------
        ntasks = omp_get_num_threads ()
        write (*,*)
        write (*,*)' juego de la vida on OMP '
        write (*,*)' number_of_processors = ', ntasks
c ------------------------------------------------------------------
c   tablero: impresion eventual
c ------------------------------------------------------------------
        if ( imprime .and. n .le. cmax ) then
          open  (1, file = 'vida_omp.map', status = 'unknown')
          close (1, status = 'delete') 
        end if
c ------------------------------------------------------------------
c   inicia cronometro;  t1 = second ()
c ------------------------------------------------------------------
        time = cero
        call system_clock (count = h1, count_rate = h3)
c ------------------------------------------------------------------
c   inicializa tablero
c ------------------------------------------------------------------
        m = n / 2
        n_vivas = 0
        tablero = 0 
C$OMP   parallel do reduction (+: n_vivas)
        do  i = m - h/2 + 1, m + h/2
        tablero (i,m) = 1
        n_vivas = n_vivas + 1
        end do ! i
c ------------------------------------------------------------------
c   ajusta cronometro
c ------------------------------------------------------------------
        call system_clock (count = h2)     ! t2 = second ()
        time (1) = h2 - h1                 ! time (1) = t2 - t1
        call system_clock (count = h1)     ! t1 = second ()
c ------------------------------------------------------------------
c   lazo sobre las generaciones
c ------------------------------------------------------------------
        do   p = 1, n_generaciones
c ------------------------------------------------------------------
c   tablero: impresion eventual
c ------------------------------------------------------------------
        if ( imprime .and. n .le. cmax ) then
          open  (1, file = 'vida_omp.map', status = 'unknown')
          do
          read  (1,*,end = 40)
          end do
   40     write (1,'(i5)') p
          do     i = 1, n
          do     j = 1, n
          if ( tablero (i,j) .eq. 1 ) then
            write (1,80,advance='no') '*'
          else
            write (1,80,advance='no') '.'
          end if
          end do ! j
          write (1,*,advance='yes')
          end do ! i
          close (1, status = 'keep') 
        end if
c ------------------------------------------------------------------
c   en las aristas impone condiciones periodicas
c ------------------------------------------------------------------
C$OMP   parallel do 
        do  i = 1, n
          tablero (i,0)   = tablero (i,n)
          tablero (i,n+1) = tablero (i,1)
          tablero (0,i)   = tablero (n,i)
          tablero (n+1,i) = tablero (1,i)
        end do ! i
c ------------------------------------------------------------------
c   calcula vecinos
c ------------------------------------------------------------------
C$OMP   parallel do 
        do  j = 1, n
        do  i = 1, n
        vecinos (i,j) = tablero (i  ,j+1) + tablero (i  ,j-1)
     #                + tablero (i+1,j  ) + tablero (i-1,j  )
     #                + tablero (i+1,j+1) + tablero (i-1,j-1)
     #                + tablero (i+1,j-1) + tablero (i-1,j+1)
        end do ! i
        end do ! j
c ------------------------------------------------------------------
c   impone reglas de juego
c ------------------------------------------------------------------
        do   j = 1, n
        do   i = 1, n
        if     ( vecinos (i,j) < 2 .or. vecinos (i,j) > 3 ) then
          tablero (i,j) = 0
        elseif ( vecinos (i,j) == 3 ) then
          tablero (i,j) = 1
        end if
        end do ! i
        end do ! j
c ------------------------------------------------------------------
c   pasa a la siguiente generacion p
c ------------------------------------------------------------------
        end do ! p
c -----------------------------------------------------------------
c   ajusta cronometro
c -----------------------------------------------------------------
        call system_clock (count = h2)     ! t2 = second ()
        time (2) = h2 - h1                 ! time (2) = t2 - t1
        time (3) = time (1) + time (2)     ! total_time
        time = time / h3                   ! pasa de ciclos a seg
c ------------------------------------------------------------------
c   calcula numero de celulas vivas remanentes
c ------------------------------------------------------------------
        n_vivas = 0
C$OMP   parallel do reduction (+: n_vivas)
        do   j = 1, n
        do   i = 1, n
        if ( tablero (i,j) == 1 ) n_vivas = n_vivas + 1
        end do ! i
        end do ! j
c ------------------------------------------------------------------
c   imprime
c ------------------------------------------------------------------
        write (*,*)
        write (*,*)' tpo de inicial.  [s] ; t1 = ', real (time(1))
        write (*,*)' tpo de barrido   [s] ; t2 = ', real (time(2))
        write (*,*)' tiempo_total     [s] ; t3 = ', real (time(3))
        write (*,*)
c ------------------------------------------------------------------
c   imprime en archivo historico
c ------------------------------------------------------------------
        open (1, file = 'vida_omp.tpo', status = 'unknown')
        do
          read (1,*, end = 50)  i
        end do
   50   write  (1,60)  ntasks, time (1:3)
        close  (1, status = 'keep')
c ------------------------------------------------------------------
c   codigos de formato
c ------------------------------------------------------------------
   60   format (1x, i4, 3 (1x, e19.9) )
   70   format (i1)
   80   format (a1)
c ..................................................................
        end program
c ==================================================================